name = "obstools_ipgp"
"""
Command-line programs for evaluating and processing OBS data

Data must be stored in SeisComp Data Structure (SDSP format
Metadata must be stored in StationXML format

command-line programs:
     obstest: Plot figures for evaluating instrumentation tests
     plotPSDs: Plot individual and all-station Power Spectral Densities
     date_extent: Coming!
     drift_correl: Coming!
     toSEGY: Coming!
"""
